@@ -195,6 +195,8 @@ _gcry_pk_map_name (const char *string)
     return 0;
   if (spec->flags.disabled)
     return 0;
+  if (!spec->flags.fips && fips_mode ())
+    return 0;
   return spec->algo;
 }
 
@@ -224,7 +226,7 @@ check_pubkey_algo (int algo, unsigned use)
   gcry_pk_spec_t *spec;
 
   spec = spec_from_algo (algo);
-  if (spec)
+  if (spec && !spec->flags.disabled && (spec->flags.fips || !fips_mode ()))
     {
       if (((use & GCRY_PK_USAGE_SIGN)
 	   && (! (spec->use & GCRY_PK_USAGE_SIGN)))
@@ -322,7 +324,11 @@ _gcry_pk_encrypt (gcry_sexp_t *r_ciph, gcry_sexp_t s_data, gcry_sexp_t s_pkey)
   if (rc)
     goto leave;
 
-  if (spec->encrypt)
+  if (spec->flags.disabled)
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (!spec->flags.fips && fips_mode ())
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (spec->encrypt)
     rc = spec->encrypt (r_ciph, s_data, keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
@@ -374,7 +380,11 @@ _gcry_pk_decrypt (gcry_sexp_t *r_plain, gcry_sexp_t s_data, gcry_sexp_t s_skey)
   if (rc)
     goto leave;
 
-  if (spec->decrypt)
+  if (spec->flags.disabled)
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (!spec->flags.fips && fips_mode ())
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (spec->decrypt)
     rc = spec->decrypt (r_plain, s_data, keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
@@ -427,7 +437,11 @@ _gcry_pk_sign (gcry_sexp_t *r_sig, gcry_sexp_t s_hash, gcry_sexp_t s_skey)
   if (rc)
     goto leave;
 
-  if (spec->sign)
+  if (spec->flags.disabled)
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (!spec->flags.fips && fips_mode ())
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (spec->sign)
     rc = spec->sign (r_sig, s_hash, keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
@@ -438,6 +452,147 @@ _gcry_pk_sign (gcry_sexp_t *r_sig, gcry_sexp_t s_hash, gcry_sexp_t s_skey)
 }
 
 
+gcry_err_code_t
+_gcry_pk_sign_md (gcry_sexp_t *r_sig, const char *tmpl, gcry_md_hd_t hd_orig,
+                  gcry_sexp_t s_skey, gcry_ctx_t ctx)
+{
+  gcry_err_code_t rc;
+  gcry_pk_spec_t *spec;
+  gcry_sexp_t keyparms = NULL;
+  gcry_sexp_t s_hash = NULL;
+  int algo;
+  const unsigned char *digest;
+  gcry_error_t err;
+  gcry_md_hd_t hd;
+  char *s;
+  char *hash_name;
+
+  *r_sig = NULL;
+
+  /* Check if it has fixed hash name or %s */
+  s = strstr (tmpl, "(hash ");
+  if (s == NULL)
+    return GPG_ERR_DIGEST_ALGO;
+
+  s += 6;
+  if (!strncmp (s, "%s", 2))
+    hash_name = NULL;
+  else
+    {
+      char *p;
+
+      for (p = s; *p && *p != ' '; p++)
+	;
+
+      hash_name = xtrymalloc (p - s + 1);
+      if (!hash_name)
+	return gpg_error_from_syserror ();
+      memcpy (hash_name, s, p - s);
+      hash_name[p - s] = 0;
+    }
+
+  err = _gcry_md_copy (&hd, hd_orig);
+  if (err)
+    {
+      xfree (hash_name);
+      return gpg_err_code (err);
+    }
+
+  if (hash_name)
+    {
+      algo = _gcry_md_map_name (hash_name);
+      if (algo == 0
+          || (fips_mode () && algo == GCRY_MD_SHA1))
+	{
+	  xfree (hash_name);
+	  _gcry_md_close (hd);
+	  return GPG_ERR_DIGEST_ALGO;
+	}
+
+      digest = _gcry_md_read (hd, algo);
+    }
+  else
+    {
+      algo = _gcry_md_get_algo (hd);
+
+      if (fips_mode () && algo == GCRY_MD_SHA1)
+        {
+          _gcry_md_close (hd);
+          return GPG_ERR_DIGEST_ALGO;
+        }
+
+      digest = _gcry_md_read (hd, 0);
+    }
+
+  if (!digest)
+    {
+      xfree (hash_name);
+      _gcry_md_close (hd);
+      return GPG_ERR_NOT_IMPLEMENTED;
+    }
+
+  if (!ctx)
+    {
+      if (hash_name)
+	rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
+			       (int) _gcry_md_get_algo_dlen (algo),
+			       digest);
+      else
+	rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
+			       _gcry_md_algo_name (algo),
+			       (int) _gcry_md_get_algo_dlen (algo),
+			       digest);
+    }
+  else
+    {
+      const unsigned char *p;
+      size_t len;
+
+      rc = _gcry_pk_get_random_override (ctx, &p, &len);
+      if (rc)
+        {
+          _gcry_md_close (hd);
+          return rc;
+        }
+
+      if (hash_name)
+	rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
+			       (int) _gcry_md_get_algo_dlen (algo),
+			       digest,
+			       (int) len, p);
+      else
+	rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
+			       _gcry_md_algo_name (algo),
+			       (int) _gcry_md_get_algo_dlen (algo),
+			       digest,
+			       (int) len, p);
+    }
+
+  xfree (hash_name);
+  _gcry_md_close (hd);
+  if (rc)
+    return rc;
+
+  rc = spec_from_sexp (s_skey, 1, &spec, &keyparms);
+  if (rc)
+    goto leave;
+
+  if (spec->flags.disabled)
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (!spec->flags.fips && fips_mode ())
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (spec->sign)
+    rc = spec->sign (r_sig, s_hash, keyparms);
+  else
+    rc = GPG_ERR_NOT_IMPLEMENTED;
+
+ leave:
+  sexp_release (s_hash);
+  sexp_release (keyparms);
+  return rc;
+}
+
+
 /*
    Verify a signature.
 
@@ -456,7 +611,11 @@ _gcry_pk_verify (gcry_sexp_t s_sig, gcry_sexp_t s_hash, gcry_sexp_t s_pkey)
   if (rc)
     goto leave;
 
-  if (spec->verify)
+  if (spec->flags.disabled)
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (!spec->flags.fips && fips_mode ())
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (spec->verify)
     rc = spec->verify (s_sig, s_hash, keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
@@ -467,6 +626,83 @@ _gcry_pk_verify (gcry_sexp_t s_sig, gcry_sexp_t s_hash, gcry_sexp_t s_pkey)
 }
 
 
+gcry_err_code_t
+_gcry_pk_verify_md (gcry_sexp_t s_sig, const char *tmpl, gcry_md_hd_t hd_orig,
+                    gcry_sexp_t s_pkey, gcry_ctx_t ctx)
+{
+  gcry_err_code_t rc;
+  gcry_pk_spec_t *spec;
+  gcry_sexp_t keyparms = NULL;
+  gcry_sexp_t s_hash = NULL;
+  int algo;
+  const unsigned char *digest;
+  gcry_error_t err;
+  gcry_md_hd_t hd;
+
+  err = _gcry_md_copy (&hd, hd_orig);
+  if (err)
+    return gpg_err_code (err);
+
+  algo = _gcry_md_get_algo (hd);
+
+  if (fips_mode () && algo == GCRY_MD_SHA1)
+    return GPG_ERR_DIGEST_ALGO;
+
+  digest = _gcry_md_read (hd, 0);
+  if (!digest)
+    {
+      _gcry_md_close (hd);
+      return GPG_ERR_DIGEST_ALGO;
+    }
+
+  if (!ctx)
+    rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
+                           _gcry_md_algo_name (algo),
+                           (int) _gcry_md_get_algo_dlen (algo),
+                           digest);
+  else
+    {
+      const unsigned char *p;
+      size_t len;
+
+      rc = _gcry_pk_get_random_override (ctx, &p, &len);
+      if (rc)
+        {
+          _gcry_md_close (hd);
+          return rc;
+        }
+
+      rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
+                             _gcry_md_algo_name (algo),
+                             (int) _gcry_md_get_algo_dlen (algo),
+                             digest,
+                             (int) len, p);
+    }
+
+  _gcry_md_close (hd);
+  if (rc)
+    return rc;
+
+  rc = spec_from_sexp (s_pkey, 0, &spec, &keyparms);
+  if (rc)
+    goto leave;
+
+  if (spec->flags.disabled)
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (!spec->flags.fips && fips_mode ())
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (spec->verify)
+    rc = spec->verify (s_sig, s_hash, keyparms);
+  else
+    rc = GPG_ERR_NOT_IMPLEMENTED;
+
+ leave:
+  sexp_release (s_hash);
+  sexp_release (keyparms);
+  return rc;
+}
+
+
 /*
    Test a key.
 
@@ -487,7 +723,11 @@ _gcry_pk_testkey (gcry_sexp_t s_key)
   if (rc)
     goto leave;
 
-  if (spec->check_secret_key)
+  if (spec->flags.disabled)
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (!spec->flags.fips && fips_mode ())
+    rc = GPG_ERR_PUBKEY_ALGO;
+  else if (spec->check_secret_key)
     rc = spec->check_secret_key (keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
@@ -569,7 +809,7 @@ _gcry_pk_genkey (gcry_sexp_t *r_key, gcry_sexp_t s_parms)
   spec = spec_from_name (name);
   xfree (name);
   name = NULL;
-  if (!spec)
+  if (!spec || spec->flags.disabled || (!spec->flags.fips && fips_mode ()))
     {
       rc = GPG_ERR_PUBKEY_ALGO; /* Unknown algorithm.  */
       goto leave;
@@ -609,6 +849,10 @@ _gcry_pk_get_nbits (gcry_sexp_t key)
 
   if (spec_from_sexp (key, 0, &spec, &parms))
     return 0; /* Error - 0 is a suitable indication for that.  */
+  if (spec->flags.disabled)
+    return 0;
+  if (!spec->flags.fips && fips_mode ())
+    return 0;
 
   nbits = spec->get_nbits (parms);
   sexp_release (parms);
@@ -742,6 +986,10 @@ _gcry_pk_get_curve (gcry_sexp_t key, int iterator, unsigned int *r_nbits)
         return NULL;
     }
 
+  if (spec->flags.disabled)
+    return NULL;
+  if (!spec->flags.fips && fips_mode ())
+    return NULL;
   if (spec->get_curve)
     result = spec->get_curve (keyparms, iterator, r_nbits);
 
@@ -927,17 +1175,6 @@ _gcry_pubkey_get_sexp (gcry_sexp_t *r_sexp, int mode, gcry_ctx_t ctx)
 gcry_err_code_t
 _gcry_pk_init (void)
 {
-  if (fips_mode())
-    {
-      /* disable algorithms that are disallowed in fips */
-      int idx;
-      gcry_pk_spec_t *spec;
-
-      for (idx = 0; (spec = pubkey_list[idx]); idx++)
-        if (!spec->flags.fips)
-          spec->flags.disabled = 1;
-    }
-
   return 0;
 }
 
@@ -952,7 +1189,9 @@ _gcry_pk_selftest (int algo, int extended, selftest_report_func_t report)
 
   algo = map_algo (algo);
   spec = spec_from_algo (algo);
-  if (spec && !spec->flags.disabled && spec->selftest)
+  if (spec && !spec->flags.disabled
+      && (spec->flags.fips || !fips_mode ())
+      && spec->selftest)
     ec = spec->selftest (algo, extended, report);
   else
     {
@@ -961,7 +1200,8 @@ _gcry_pk_selftest (int algo, int extended, selftest_report_func_t report)
          of an encryption mode (e.g. pkcs1, ecdsa, or ecdh).  */
       if (report)
         report ("pubkey", algo, "module",
-                spec && !spec->flags.disabled?
+                spec && !spec->flags.disabled
+                && (spec->flags.fips || !fips_mode ())?
                 "no selftest available" :
                 spec? "algorithm disabled" :
                 "algorithm not found");
@@ -969,3 +1209,48 @@ _gcry_pk_selftest (int algo, int extended, selftest_report_func_t report)
 
   return gpg_error (ec);
 }
+
+
+struct pk_random_override {
+  size_t len;
+  unsigned char area[];
+};
+
+gpg_err_code_t
+_gcry_pk_random_override_new (gcry_ctx_t *r_ctx,
+                              const unsigned char *p, size_t len)
+{
+  gcry_ctx_t ctx;
+  struct pk_random_override *pro;
+
+  *r_ctx = NULL;
+  if (!p)
+    return GPG_ERR_EINVAL;
+
+  ctx = _gcry_ctx_alloc (CONTEXT_TYPE_RANDOM_OVERRIDE,
+                         sizeof (size_t) + len, NULL);
+  if (!ctx)
+    return gpg_err_code_from_syserror ();
+  pro = _gcry_ctx_get_pointer (ctx, CONTEXT_TYPE_RANDOM_OVERRIDE);
+  pro->len = len;
+  memcpy (pro->area, p, len);
+
+  *r_ctx = ctx;
+  return 0;
+}
+
+gpg_err_code_t
+_gcry_pk_get_random_override (gcry_ctx_t ctx,
+                              const unsigned char **r_p, size_t *r_len)
+{
+  struct pk_random_override *pro;
+
+  pro = _gcry_ctx_find_pointer (ctx, CONTEXT_TYPE_RANDOM_OVERRIDE);
+  if (!pro)
+    return GPG_ERR_EINVAL;
+
+  *r_p = pro->area;
+  *r_len = pro->len;
+
+  return 0;
+}
